unit ubiomasaembalsable;

{$mode delphi}


interface

uses
  Math,
  SysUtils, Classes, xmatdefs, uGTer, uNodos,
  uglobs,
  usimplex,
  ufichasLPD,
  ufechas,
  ucosa, uranddispos,
  uconstantesSimSEE,
  uFuentesAleatorias, uActores, uEstados;

resourcestring
  rsGeneradorBiomasaEmbalsable =
    'Generador de Biomasa Embalsable';


type
  TFichaBiomasaEmbalsable = class(TFichaLPD)
  public
    PMax: NReal;// MW máximos de una unidad generadora
    PEC: NReal; // [USD/MWh] Pago por Energía Comprometida
    PEE: NReal; // [USD/MWh] Pago por Energía Entregada (o vertida)
    fd: NReal;  // factor de disponibilidad forturita
    TMR: NReal; // Tiempo Medio de Reparción (horas)
    cvea_impuesto: NReal; // [USD/MWh] valor de la energía almacenada
    // si flg_imponer_cvea = true.

    constructor Create(capa: integer; fecha: TFecha; periodicidad: TPeriodicidad;
      fd, TMR: NReal; PMax: NReal; PEC, PEE, cvea_impuesto: NReal);
    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    procedure generarLineaResumen(var archi: TextFile); override;
    function infoAd_: string; override;
    procedure Free; override;
  end;

  TBiomasaEmbalsable = class(TGTer)
  public

    // parametros persistentes
    X_EA_inicial: NReal;  // Valor del estado al inicio.
    ECA: NReal; // Energía comprometida ANUAL
    NPuntos_ECA: integer; // Cantidad de puntos de discretización de X_ECA
    ixPEC_fuente: TFuenteAleatoria;
    ixPEC_Borne_nombre: string;
    ixPEC_Borne_numero: integer;
    ixPEE_fuente: TFuenteAleatoria;
    ixPEE_Borne_nombre: string;
    ixPEE_Borne_numero: integer;
    flg_imponer_cvea: boolean;

    pa: TFichaBiomasaEmbalsable;
    PEC, PEE: NReal; // USD/MWh indexados.
    cvea_impuesto: NReal; // USD/MWh indexado.

    ecp: NReal; // energía comprometida del paso.
    edp: NReal; // energía despachable del paso. PN * HorasDelPaso

    // Variable de estado  Energia Almaceanda MWh
    X_EA: NReal;
    Xs_EA: NReal;

    PNominalPorUnidad: NReal; // es pa.PMax salvo que sea sin sorteos
    NMaquinasDisponibles: integer; // maximo de acople
    NMaquinasDespachables: integer; // Para cuantas da la energía del paso

    // -dCF/dX
    cvea: NReal;

    // REsultados del paso
    NMaquinasDespachadas: integer;
    EnergiaVertida, EnergiaDespachada: NReal;

    constructor Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
      lpdUnidades, lpd: TFichasLPD; nodo: TNodo; ECA: NReal;
      NPuntosECA: integer; X_EA_Inicial: NReal; ixPEC_fuente: TFuenteAleatoria;
      ixPEC_Borne_nombre: string; ixPEE_fuente: TFuenteAleatoria;
      ixPEE_Borne_nombre: string; imponer_cvea: boolean; TonCO2xMWh: NReal;
      LowCostMustRun, CleanDevelopmentMechanism: boolean);
      reintroduce;

    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    procedure PrepararMemoria(globs: TGlobs); override;
    procedure RegistrarParametrosDinamicos; override;

    function PotenciaFirme: NReal; override;

    function InfoAd_: string; override;
    class function DescClase: string; override;

    procedure PosicionarseEnEstrellita; override;
    procedure ActualizarEstadoGlobal; override;

    procedure Sim_Cronica_Inicio; override;
    procedure SorteosDelPaso(sortear: boolean); override;
    procedure PrepararPaso_ps; override;

    procedure EvolucionarEstado; override;

    //      procedure Sim_Cronica_Fin; override;

    procedure opt_nvers(var ivar, ivae, ires: integer); override;
    procedure optx_nvxs(var ixr, ixd, iauxNReal, iauxInt: integer); override;
    procedure optx_RegistrarVariablesDeEstado(adminEstados: TAdminEstados);
      override;

    procedure opt_cargue(s: TSimplex); override;
    procedure opt_fijarRestriccionesDeCaja(s: TSimplex); override;
    procedure opt_leerSolucion(s: TSimplex); override;
{$IFDEF SPXCONLOG}
    procedure spx_NombrarVariables(s: TSimplex); override;
{$ENDIF}
    function getNombreVar(ivar: integer; var nombre: string): boolean; override;
    function getNombreRes(ires: integer; var nombre: string): boolean; override;

    procedure Free; override;

    procedure PubliVars; override;
    procedure dump_Variables(var f: TextFile; charIndentacion: char); override;

    class function TipoFichaLPD: TClaseDeFichaLPD; override;
  end;

procedure cambioFichaPD_TBiomasaEmbalsable(actor: TActor);
procedure cambioFichaUnidades_TBiomasaEmbalsable(actor: TActor);

procedure AlInicio;
procedure AlFinal;

implementation


//----------------------------------
// Métodos de TFichaBiomasaEmbalsable
//==================================
constructor TFichaBiomasaEmbalsable.Create(capa: integer; fecha: TFecha;
  periodicidad: TPeriodicidad; fd, TMR: NReal; PMax: NReal;
  PEC, PEE, cvea_impuesto: NReal);
begin
  inherited Create(capa, fecha, periodicidad);
  self.fd := fd;
  self.TMR := TMR;
  self.PMax := PMax;
  self.PEC := PEC;
  self.PEE := PEE;
  self.cvea_impuesto := cvea_impuesto;
end;

constructor TFichaBiomasaEmbalsable.Create_ReadFromText(f: TArchiTexto);
begin

  inherited Create_ReadFromText(f);

  f.IniciarLecturaRetrasada;
  f.rd('fd', fd);
  f.rd('TMR', TMR);
  f.rd('PMax', PMax);
  f.rd('PEC', PEC);
  f.rd('PEE', PEE);
  if f.version > 106 then
    f.rd('cvea_impuesto', cvea_impuesto)
  else
    cvea_impuesto := 0;

  f.EjecutarLectura;
end;

procedure TFichaBiomasaEmbalsable.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wr('fd', fd);
  f.wr('TMR', TMR);
  f.wr('PMax', PMax);
  f.wr('PEC', PEC);
  f.wr('PEE', PEE);
  f.wr('cvea_impuesto', cvea_impuesto);
end;

procedure TFichaBiomasaEmbalsable.generarLineaResumen(var archi: TextFile);
begin
  Write(archi, '-', #9,       //PMín
    FloatToStrF(PMax, formatoReales, 8, 1), #9,  //PMáx
    '-', #9,       //CV_Mín
    FloatToStrF(PEC + PEE, formatoReales, 8, 2), #9,    //CV_Medio
    FloatToStrF(PEE, formatoReales, 8, 2), #9,    //CV_Incremental
    FloatToStrF(fd, formatoReales, 8, 2), #9,  //FDisp
    '-', #9,       //Costo Arranque
    '-', #9,       //Costo Parada
    '-', #9,       //mínNPasosOn
    '-', #9,       //mínNPasosOff
    '-', #9,       //desiciónPasosOnPorCiclo
    '-', #9,       //desiciónPasosOffPorCiclo
    '-', #9,       //costoPorCicloOn
    '-', #9);      //costoPorCicloOff
end;

function TFichaBiomasaEmbalsable.InfoAd_: string;
begin
  Result := 'PMáx= ' + FloatToStrF(PMax, ffGeneral, 10, 1) + ' MW, ' +
    'PE= ' + FloatToStrF(PEC + PEE, ffGeneral, 10, 1) + ' USD/MWh, ' +
    'fd= ' + FloatToStrF(fd, ffGeneral, 10, 2) + ' p.u., ' + 'TMR= ' +
    FloatToStrF(TMR, ffGeneral, 10, 1) + ' hs';
end;

procedure TFichaBiomasaEmbalsable.Free;
begin
  inherited Free;
end;

//-----------------------------
// Métodos de TBiomasaEmbalsable
//=============================
procedure TBiomasaEmbalsable.dump_Variables(var f: TextFile; charIndentacion: char);
begin
  inherited dump_Variables(f, charIndentacion);
  writeln(f, charIndentacion, 'PEC[USD/MWh]= ', FloatToStrF(PEC, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'PMAx[MW]= ', FloatToStrF(pa.PMax, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'X_EA= ', X_EA);
  writeln(f);
end;

class function TBiomasaEmbalsable.TipoFichaLPD: TClaseDeFichaLPD;
begin
  Result := TFichaBiomasaEmbalsable;
end;

constructor TBiomasaEmbalsable.Create(capa: integer; nombre: string;
  nacimiento, muerte: TFecha; lpdUnidades, lpd: TFichasLPD; nodo: TNodo;
  ECA: NReal; NPuntosECA: integer; X_EA_Inicial: NReal;
  ixPEC_fuente: TFuenteAleatoria; ixPEC_Borne_nombre: string;
  ixPEE_fuente: TFuenteAleatoria; ixPEE_Borne_nombre: string;
  imponer_cvea: boolean; TonCO2xMWh: NReal;
  LowCostMustRun, CleanDevelopmentMechanism: boolean);
begin
  inherited Create(capa, nombre, nacimiento, muerte, lpdUnidades, lpd, nodo,
    TonCO2xMWh, LowCostMustRun, CleanDevelopmentMechanism);
  P := nil;
  self.ECA := ECA;
  self.NPuntos_ECA := NPuntos_ECA;
  self.X_EA_Inicial := X_EA_Inicial;
  self.ixPEC_fuente := ixPEC_fuente;
  self.ixPEC_Borne_nombre := ixPEC_Borne_nombre;
  self.ixPEC_Borne_numero := -1;
  self.ixPEE_fuente := ixPEE_fuente;
  self.ixPEE_Borne_nombre := ixPEE_Borne_nombre;
  self.ixPEE_Borne_numero := -1;
  self.flg_imponer_cvea := imponer_cvea;
end;

constructor TBiomasaEmbalsable.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create_ReadFromText(f);

  f.rd('ECA', ECA);
  f.rd('NPuntos_ECA', NPuntos_ECA);
  f.rd('X_EA_inicial', X_EA_inicial);

  f.rdReferencia('ixPEC_fuente', TCosa(ixPEC_fuente), self);
  f.rd('ixPEC_Borne_nombre', ixPEC_Borne_nombre);
  f.rdReferencia('ixPEE_fuente', TCosa(ixPEC_fuente), self);
  f.rd('ixPEE_Borne_nombre', ixPEC_Borne_nombre);
  if f.Version > 106 then
    f.rd('flg_imponer_cvea', flg_imponer_cvea)
  else
    flg_imponer_cvea := False;

  P := nil;
  pa := nil;
  nodo := nil;
end;

procedure TBiomasaEmbalsable.prepararMemoria(globs: TGlobs);
begin
  inherited prepararMemoria(globs);
  if ixPEC_fuente <> nil then
    ixPEC_Borne_numero := ixPEC_fuente.IdBorne(ixPEC_Borne_nombre);
  if ixPEE_fuente <> nil then
    ixPEE_Borne_numero := ixPEE_fuente.IdBorne(ixPEE_Borne_nombre);
end;

procedure TBiomasaEmbalsable.registrarParametrosDinamicos;
begin
  uActores.procCambioFichaUnidades := @cambioFichaUnidades_TBiomasaEmbalsable;
  inherited registrarParametrosDinamicos;
  uActores.procCambioFichaUnidades := nil;
  lpd.expandirFichas(globs);
  lpd.RegistrarFichasAActualizar(Self, globs.ActualizadorLPD, @pA,
    nil, @cambioFichaPD_TBiomasaEmbalsable);
end;

function TBiomasaEmbalsable.potenciaFirme: NReal;
begin
  Result := (paUnidades.nUnidades[0]) * pa.PMax * pa.fd;
end;

function TBiomasaEmbalsable.infoAd_: string;
begin
  Result := '';
end;

class function TBiomasaEmbalsable.DescClase: string;
begin
  Result := rsGeneradorBiomasaEmbalsable;
end;

procedure TBiomasaEmbalsable.PosicionarseEnEstrellita;
begin
  if not flg_imponer_cvea then
    X_EA := globs.CF.xr[ixr]
  else
    X_EA := ECA / 2.0;
end;

procedure TBiomasaEmbalsable.ActualizarEstadoGlobal;
begin
  if not flg_imponer_cvea then
    globs.CF.xr[ixr] := X_EA;
end;

procedure TBiomasaEmbalsable.Sim_Cronica_Inicio;
begin
  inherited Sim_Cronica_Inicio;
  X_EA := X_EA_inicial;
end;

procedure TBiomasaEmbalsable.SorteosDelPaso(sortear: boolean);
begin
  if hayForzamientos or globs.ObligarDisponibilidad_1_ then
  begin
    NMaquinasDisponibles := paUnidades.nUnidades[0];
    PNominalPorUnidad := pa.PMax;
  end
  else if sortear then
  begin
    ActualizarProbabilidadesReparacionYRotura_(pa.fd, pa.TMR);
    NMaquinasDisponibles := Sorteos_RepRotUnidades;
    PNominalPorUnidad := pa.PMax;
  end
  else
  begin
    NMaquinasDisponibles := paUnidades.nUnidades[0];
    PNominalPorUnidad := pa.PMax * pa.fd;
  end;
end;

procedure TBiomasaEmbalsable.prepararPaso_ps;
var
  dCFdX_inc, dCFdX_dec: NReal;
  rescod: integer;
  xrpos: NReal;
  dCFdX: NReal;
begin
  // calculamos los valores PEC y PEE en base los precios de la ficha
  // de parámetros dinámicos activa y los valores de las fuentes de
  // indexación de ambos precios.
  if ixPEC_fuente <> nil then
    PEC := pa.PEC * ixPEC_fuente.bornera[ixPEC_Borne_numero]
  else
    PEC := pa.PEC;

  if ixPEE_fuente <> nil then
  begin
    PEE := pa.PEE * ixPEE_fuente.bornera[ixPEE_Borne_numero];
    if flg_imponer_cvea then
      cvea_impuesto := pa.cvea_impuesto * ixPEE_fuente.bornera[ixPEE_Borne_numero];
  end
  else
  begin
    PEE := pa.PEE;
    if flg_imponer_cvea then
      cvea_impuesto := pa.cvea_impuesto;
  end;



  if not flg_imponer_cvea then
  begin
    if globs.EstadoDeLaSala = CES_OPTIMIZANDO then
      globs.CF.devxr_estrella(ixr, globs.kPaso_ + 1, dCFdX_inc, dCFdX_dec, rescod)
    else
      globs.CF.devxr_continuo(ixr, globs.kPaso_ + 1, dCFdX_inc,
        dCFdX_dec, rescod, xrpos);

    case rescod of
      -1: dCFdX := dCFdX_inc;
      0: dCFdX := (dCFdX_dec + dCFdX_inc) / 2.0;
      1: dCFdX := dCFdX_dec;
    end;

    cvea := -dCFdX * globs.fActPaso;
  end
  else
    cvea := cvea_impuesto;

  if NMaquinasDisponibles = 0 then
    NMaquinasDespachables := 0
  else
  begin
    NMaquinasDespachables := trunc((X_EA + ecp) / edp);
    NMaquinasDespachables := min(NMaquinasDisponibles, NMaquinasDespachables);
  end;

end;

procedure TBiomasaEmbalsable.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wr('ECA', ECA);
  f.wr('NPuntos_ECA', NPuntos_ECA);
  f.wr('X_EA_inicial', X_EA_inicial);

  f.wrReferencia('ixPEC_fuente', TCosa(ixPEC_fuente));
  f.wr('ixPEC_Borne_nombre', ixPEC_Borne_nombre);
  f.wrReferencia('ixPEE_fuente', TCosa(ixPEC_fuente));
  f.wr('ixPEE_Borne_nombre', ixPEC_Borne_nombre);

  f.wr('flg_imponer_cvea', flg_imponer_cvea);
end;

procedure TBiomasaEmbalsable.opt_nvers(var ivar, ivae, ires: integer);
begin
  if NMaquinasDespachables = 0 then
    exit; // si no hay máquinas no juego

  Self.ivar := ivar;
  ivar := ivar + 3; // A, VE, Xs_EA

  self.ivae := ivae;
  Inc(ivae);

  Self.ires := ires;
  Inc(ires);  //   - A* edp -VE -Xs_EA + x_EA + ecp = 0
end;

procedure TBiomasaEmbalsable.optx_nvxs(var ixr, ixd, iauxNReal, iauxInt: integer);
begin
  if not flg_imponer_cvea then
  begin
    Self.ixr := ixr;
    ixr := ixr + 1;
  end;
end;

procedure TBiomasaEmbalsable.optx_RegistrarVariablesDeEstado(
  adminEstados: TAdminEstados);
begin
  if not flg_imponer_cvea then
    adminEstados.Registrar_Continua(ixr, 0, ECA, NPuntos_ECA, 'ECA', 'MWh');
end;

procedure TBiomasaEmbalsable.opt_cargue(s: TSimplex);
var
  inodores: integer;
  iposte: integer;
begin
  if NMaquinasDespachables = 0 then
    exit; // si no hay máquinas no juego

{$IFDEF SPXCONLOG}
  spx_NombrarVariables(s);
{$ENDIF}

  inodores := nodo.ires;
  for iposte := 0 to globs.NPostes - 1 do
  begin
    // cargamos la potencia en la restricción de balance de potencia del nodo
    s.pon_e(inodores + iposte, ivar, PNominalPorUnidad);
  end;

  // Cargamos la restricción dinámica
  // - A* edp -VE -Xs_EA + x_EA + ecp = 0
  s.pon_e(ires, ivar, -edp);
  s.pon_e(ires, ivar + 1, -1);
  s.pon_e(ires, ivar + 2, -1);
  s.pon_e(ires, s.nc, x_EA + ecp);

  // Ahora agregamos a la función de UTILIDAD (-costo
  s.pon_e(s.nf, ivar, -edp * PEE); // - A * edp  * PEE
  s.pon_e(s.nf, ivar + 1, -PEE); // - VE * PEE
  s.pon_e(s.nf, ivar + 2, cvea);  // - cvea * (X - Xs )
  s.acum_e(s.nf, s.nc, -cvea * X_EA - PEC * ecp);

end;

procedure TBiomasaEmbalsable.opt_fijarRestriccionesDeCaja(s: TSimplex);
var
  iposte: integer;
  AForzadas: integer;

begin
  if NMaquinasDespachables = 0 then
    exit; // si no hay máquinas no juego


  if (not hayForzamientos) then
  begin
    s.cota_sup_set(ivar, NMaquinasDespachables);
  end
  else
  begin
    AForzadas := trunc(paForzamiento.P[0] / pa.PMax + 0.5);
    s.FijarVariable(ivar, AForzadas);
  end;

  s.set_entera(ivae, ivar, NMaquinasDespachables);

  s.cota_sup_set(ivar + 1, max(0, X_EA + ecp - ECA));
  // No puede verter más que lo que le sobra si no despacha nada
  s.cota_sup_set(ivar + 2, ECA); // el estado no puede superar ECA

  s.FijarRestriccionIgualdad(ires);
end;

procedure TBiomasaEmbalsable.opt_leerSolucion(s: TSimplex);
var
  iposte: integer;
  m: NReal;

begin
  Ingreso_PorEnergia := 0;
  Ingreso_PorDisponibilidad := ecp * PEC;


  CostoDirectoDelPaso := Ingreso_PorDisponibilidad;
  if NMaquinasDespachables = 0 then
  begin
    vclear(P);
    NMaquinasDespachadas := 0;
    EnergiaDespachada := 0;
    Xs_EA := X_EA + ecp;
    if (Xs_EA > ECA) then
    begin
      EnergiaVertida := Xs_EA - ECA;
      Xs_EA := ECA;
      Ingreso_PorEnergia := EnergiaVertida * PEE;
      costoDirectoDelPaso := costoDirectoDelPaso + Ingreso_PorEnergia;
    end
    else
      EnergiaVertida := 0;
    exit;
  end;

  NMaquinasDespachadas := trunc(s.xval(ivar) + 0.1);
  EnergiaVertida := s.xval(ivar + 1);
  Xs_EA := s.xval(ivar + 2);


  P[0] := NMaquinasDespachadas * PNominalPorUnidad;
  EnergiaDespachada := NMaquinasDespachadas * edp;


  // recuperamos los valores de Potencia despachada
  for iposte := 1 to globs.NPostes - 1 do
    P[iposte] := P[0];

  Ingreso_PorEnergia := (EnergiaDespachada + EnergiaVertida) * PEE;
  costoDirectoDelPaso := costoDirectoDelPaso + Ingreso_PorEnergia;
end;

procedure TBiomasaEmbalsable.EvolucionarEstado;
begin
  X_EA := Xs_EA;
end;

{$IFDEF SPXCONLOG}
procedure TBiomasaEmbalsable.spx_NombrarVariables(s: TSimplex);
var
  iposte: integer;
begin
  if NMaquinasDespachables = 0 then
    exit;
  s.set_NombreVar(ivar, Nombre + '_Acopladas');
  s.set_NombreVar(ivar + 1, Nombre + '_EnergiaVertida');
  s.set_NombreVar(ivar + 2, Nombre + '_Xs_EA');
  s.set_NombreRest(ires, nombre + 'dinámica(EA)');
end;

{$ENDIF}

function TBiomasaEmbalsable.getNombreVar(ivar: integer; var nombre: string): boolean;
var
  kvar: integer;
  res: boolean;
begin
  res := True;
  kvar := ivar - self.ivar;
  case kvar of
    0: nombre := self.Nombre + '_Acopladas';
    1: nombre := self.Nombre + '_EnergiaVertida';
    2: nombre := self.Nombre + '_Xs_EA';
    else
      Res := False;
  end;
  Result := res;
end;

function TBiomasaEmbalsable.getNombreRes(ires: integer; var nombre: string): boolean;
begin
  if ires = self.ires then
  begin
    nombre := 'dinamica(EA)';
    Result := True;
  end
  else
    Result := False;

end;

procedure TBiomasaEmbalsable.PubliVars;
begin
  inherited PubliVars;

  PublicarVariableNR('X_EA', 'MW', 6, 1, X_EA, True);
  PublicarVariableNI('NMaquinasDespachadas', 'u', NMaquinasDespachadas, True);
  PublicarVariableNR('EnergiaVertida', 'MWh', 6, 1, EnergiaVertida, True);
  PublicarVariableNR('EnergiaDespachada', 'MWh', 6, 1, EnergiaDespachada, True);
  PublicarVariableNR('cvea', 'USD/MWh', 6, 3, cvea, True);
  PublicarVariableNR('IngresoPEC', 'USD', 6, 1, Ingreso_PorDisponibilidad, True);
  PublicarVariableNR('IngresoPEE', 'USD', 6, 1, Ingreso_PorEnergia, True);

end;

procedure TBiomasaEmbalsable.Free;
begin
  inherited Free;
end;

procedure cambioFichaPD_TBiomasaEmbalsable(actor: TActor);
begin
  with actor as TBiomasaEmbalsable do
  begin
    ecp := ECA / HorasDelAnio * globs.HorasDelPaso;
    edp := pa.PMax * globs.HorasDelPaso;
  end;
end;

procedure cambioFichaUnidades_TBiomasaEmbalsable(actor: TActor);
begin
end;

procedure AlInicio;
begin
  registrarClaseDeCosa(TBiomasaEmbalsable.ClassName, TBiomasaEmbalsable);
  registrarClaseDeCosa(TFichaBiomasaEmbalsable.ClassName, TFichaBiomasaEmbalsable);
end;

procedure AlFinal;
begin
end;

end.
